/*  Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */


#include <stdio.h>
#include <math.h>
#include <string.h>
#include <stdarg.h>

#include "mex.h"
#include <restbus/cs_dbc.h>
#include <restbus/cs_message.h>
#include <restbus/cs_signal.h>
#include <restbus/cs_debug.h>


/* Input Arguments */

#define	DBC_IN	        prhs[0]
#define	CANID_IN	prhs[1]

/* Output Arguments */

#define	STRUCT_OUT	plhs[0]
#define MSTRUCT_OUT     plhs[1]
#define ERR_OUT         plhs[2] 

#define NFIELDS 10
#define NMFIELDS 5

#define ERRORID "CANRESTBUS:Error"

/*Errortable*/
#define ERROR_NOERROR 0
#define ERROR_ARGUMENTS 1
#define ERROR_FILENAME 2
#define ERROR_FILENOTFOUND 3
#define ERROR_CANID 4
#define ERROR_SIGNAL 5
#define ERROR_MALLOC 6
#define ERROR_MESSAGE 7


#if !defined(MAX)
#define	MAX(A, B)	((A) > (B) ? (A) : (B))
#endif

#if !defined(MIN)
#define	MIN(A, B)	((A) < (B) ? (A) : (B))
#endif

void mexFunction( int nlhs, mxArray *plhs[], 
		  int nrhs, const mxArray*prhs[] )
     
{
  char func[] = "mex_can_dbc";
  cs_dbc_t * dbcfile = NULL;
  char *dbcfilename = NULL; 
  FILE *fd = NULL;
  double canid;
  mwSize m,n; 
  mwSize buflen; 
  cs_signal_t **info = NULL;
  cs_message_t *minfo = NULL;
  
  const char *field_names[] = {"bit_start",
			       "bit_len",
			       "endianess",
			       "signedness",
			       "scale",
			       "offset",
			       "min",
			       "max",
			       "name",
			       "unit"
  };


  const char *mfield_names[] = {"name",
				"id",
			       "dlc",
			       "period_s",
			       "offset_s"
  };

  
  int i;
  mwSize dimserror[2] = {1, 0};
  mwSize mdimserror[2] = {1, 1};
  double errcode = ERROR_NOERROR;
  
  int nosignals;
  mwSize dims[2];
  mwSize mdims[2];


  /* Check for proper number of arguments */
  if (nrhs != 2) { 
    mexWarnMsgIdAndTxt(ERRORID,"%s: Two input arguments required.",func); 
    errcode = ERROR_ARGUMENTS;
    goto errorhandling;
  } else if (nlhs != 3) {
    mexWarnMsgIdAndTxt(ERRORID,"%s: Three output arguments required.",func); 
    errcode = ERROR_ARGUMENTS;
    goto errorhandling;
  } 
    
  /* Check the dimensions of CANID. */ 
  m = mxGetM(CANID_IN); 
  n = mxGetN(CANID_IN);
  if (!mxIsDouble(CANID_IN) || mxIsComplex(CANID_IN) || 
      (MAX(m,n) != 1) || (MIN(m,n) != 1)) { 
    mexWarnMsgIdAndTxt(ERRORID,"%s: Second argument has to be a scalar .",func); 
    errcode = ERROR_ARGUMENTS;
    goto errorhandling;
  } 

  /* Get the dbc filename*/
  if (!mxIsChar(DBC_IN) && mxIsNumeric(DBC_IN)) { 
    mexWarnMsgIdAndTxt(ERRORID,"%s: First argument has be a filename.",func); 
    errcode = ERROR_ARGUMENTS;
    goto errorhandling;
  } 


  /*Process the DBC File*/

  /* Allocate enough memory to hold the converted string. */ 
  buflen = mxGetNumberOfElements(DBC_IN) + 1;
  dbcfilename = (char*)mxCalloc(buflen, sizeof(char));

  /* Copy the string data from string_array_ptr and place it into buf. */ 
  if (mxGetString(DBC_IN, dbcfilename, buflen) != 0) {
    mexWarnMsgIdAndTxt(ERRORID,"%s: Could not convert string data.",func);
    errcode = ERROR_ARGUMENTS;
    goto errorhandling;
  }

  /*Get CAN-ID*/
  canid = mxGetScalar(CANID_IN);

  if(dbcfilename == NULL){
    mexWarnMsgIdAndTxt(ERRORID,"%s: No valid DBC-File Name",func);
    errcode = ERROR_FILENAME;
    goto errorhandling;
  }
  
  /*Try a dummy open, if file really exists*/
  if((fd = fopen(dbcfilename,"r")) == NULL){
    mexWarnMsgIdAndTxt(ERRORID,"%s: DBC-File does not exist",func);  
    errcode = ERROR_FILENOTFOUND;
    goto errorhandling;
  }
  fclose(fd);

  if((dbcfile = cs_dbc_init()) == NULL) {
    mexWarnMsgIdAndTxt(ERRORID,"%s: Failed to init DBC structure",func);
    errcode = ERROR_FILENAME;
    goto errorhandling;
  }
  
  if(cs_dbc_read(dbcfile,dbcfilename)){
    mexWarnMsgIdAndTxt(ERRORID,"%s: No valid DBC-File",func);
    errcode = ERROR_FILENAME;
    goto errorhandling;
  }

  
   if((minfo = cs_dbc_getmessage(dbcfile,canid)) == NULL){
    mexWarnMsgIdAndTxt(ERRORID,"%s: Cannot find CAN-ID",func);
    errcode = ERROR_CANID;
    goto errorhandling;
  }

  
  nosignals = cs_message_get_no_signals(minfo);
  if(nosignals < 1){
    mexWarnMsgIdAndTxt(ERRORID,"%s: No valid signal defined",func);
    errcode = ERROR_SIGNAL;
    goto errorhandling;
  }

  if((info = (cs_signal_t**)mxCalloc(nosignals, sizeof(cs_signal_t*)))==NULL){
    mexWarnMsgIdAndTxt(ERRORID,"%s: Malloc failure",func);
    errcode = ERROR_MALLOC;
    goto errorhandling;
  }

  for(i = 0; i < nosignals; i++){
    info[i] = cs_message_get_signal_iter(minfo,i);
    if(info[i] == NULL){
      mexWarnMsgIdAndTxt(ERRORID,"%s: Failed to get signal description",func);
      errcode = ERROR_SIGNAL;
      goto errorhandling;
    }
  } 

  /*signals*/
  dims[0] = 1;
  dims[1] = nosignals;

  /*message*/
  mdims[0] = 1;
  mdims[1] = 1;

  STRUCT_OUT = mxCreateStructArray(2, dims, NFIELDS, field_names);
  MSTRUCT_OUT = mxCreateStructArray(2, mdims, NMFIELDS, mfield_names);
  ERR_OUT = mxCreateDoubleScalar(ERROR_NOERROR); /*No error*/

  /*Fill signal structure*/
  for(i=0; i < nosignals; i++) {
    mxSetField(STRUCT_OUT,i,"bit_start",mxCreateDoubleScalar((double)cs_signal_get_bitoffset(info[i])));
    mxSetField(STRUCT_OUT,i,"bit_len",mxCreateDoubleScalar((double)cs_signal_get_bitlength(info[i])));
    mxSetField(STRUCT_OUT,i,"endianess",mxCreateDoubleScalar((double)cs_signal_get_byteorder(info[i])));
    mxSetField(STRUCT_OUT,i,"signedness",mxCreateDoubleScalar((double) cs_signal_get_signal_type(info[i])));
    mxSetField(STRUCT_OUT,i,"scale",mxCreateDoubleScalar((double)cs_signal_get_scale(info[i])));
    mxSetField(STRUCT_OUT,i,"offset",mxCreateDoubleScalar((double)cs_signal_get_offset(info[i])));
    mxSetField(STRUCT_OUT,i,"min",mxCreateDoubleScalar((double)cs_signal_get_min(info[i])));
    mxSetField(STRUCT_OUT,i,"max",mxCreateDoubleScalar((double)cs_signal_get_max(info[i])));
    mxSetField(STRUCT_OUT,i,"name",mxCreateString(cs_signal_get_name(info[i])));
    mxSetField(STRUCT_OUT,i,"unit",mxCreateString(cs_signal_get_unit(info[i])));
  }

  /*Fill message structure*/
  mxSetField(MSTRUCT_OUT,0,"name",mxCreateString(cs_message_get_name(minfo)));
  mxSetField(MSTRUCT_OUT,0,"id",mxCreateDoubleScalar((double)cs_message_get_id(minfo)));
  mxSetField(MSTRUCT_OUT,0,"dlc",mxCreateDoubleScalar((double)cs_message_get_dlc(minfo)));
  mxSetField(MSTRUCT_OUT,0,"period_s",mxCreateDoubleScalar(((double)cs_message_get_interval(minfo))/1.0e6));
  mxSetField(MSTRUCT_OUT,0,"offset_s",mxCreateDoubleScalar(((double)cs_message_get_offset(minfo))/1.0e6));

  return;

 errorhandling:

  
  if(dbcfile != NULL){
    cs_dbc_free(dbcfile);
  }
  if(info != NULL){
    mxFree(info);
  }
 
  STRUCT_OUT = mxCreateStructArray(2, dimserror, NFIELDS, field_names);
  MSTRUCT_OUT = mxCreateStructArray(2, mdimserror, NMFIELDS, mfield_names);
  mxSetField(MSTRUCT_OUT,0,"name",mxCreateString("Unknown"));
  mxSetField(MSTRUCT_OUT,0,"id",mxCreateDoubleScalar(0));
  mxSetField(MSTRUCT_OUT,0,"dlc",mxCreateDoubleScalar(0));
  mxSetField(MSTRUCT_OUT,0,"period_s",mxCreateDoubleScalar(0.0));
  mxSetField(MSTRUCT_OUT,0,"offset_s",mxCreateDoubleScalar(0.0));
  ERR_OUT = mxCreateDoubleScalar(errcode);
  return;

}
