/*  Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#ifndef H_SFUN_CANRESTBUS
#define H_SFUN_CANRESTBUS

#include <restbus/cs_signal.h>

/*Only CAN 2.0 is actually supported
  Max byte length 8 byte*/
#define MAXDLC 8

#define CHECK_STRUCT_VALUE(SRC,IDX,FIELD,VALUE) if(!get_numeric_value(SRC, IDX, FIELD, VALUE)) { \
        ssPrintf("Structure failure for field: %s.\n",FIELD);           \
        ssSetErrorStatus(S,"Signalinfo structure failure.");            \
        return;                                                         \
    }

#define CHECK_STRUCT_VALUE_RVALUE(SRC,IDX,FIELD,VALUE) if(!get_numeric_value(SRC, IDX, FIELD, VALUE)) { \
        ssPrintf("Structure failure for field: %s.\n",FIELD);           \
        ssSetErrorStatus(S,"Signalinfo structure failure.");            \
        return 1;                                                       \
    }

#ifdef __cplusplus
extern "C"
{
#endif
  
    int get_numeric_value(const  mxArray *src, int idx, char_T *field, double *value)
    {
        mxArray * datainfo = NULL; 
        if((datainfo= mxGetField(src, idx, field))!=NULL) {
            *value = mxGetScalar(datainfo);
            return 1;
        }
        else {
            return 0;
        }
    }
      
    /* Function: WriteMatlabStructureToRTWFile =====================================
     *
     * ml2rtw.c : Contains utility function to write a ML structure to rtw file.
     *
     * Author(s): Murali Yeddanapudi, 13-Nov-1997
     *
     * Copyright 1990-2002 The MathWorks, Inc.
     *
     * Abstract:
     *      This routine writes out a matlab structure to the the rtw file. Nested
     *      structures inside structures are also supported, i.e., any field(s) of
     *      the mxArray passed in can be structures.  However, the basic atomic
     *      units have to be either row strings or non-sparse, numeric vectors or
     *      matrices. For example, fields cannot be things like a matrix of strings
     *      or nd arrays.
     *
     *      In order to avoid name clashes with records (espeically ones that might
     *      be added in future) written out by Simulink, your structure name should
     *      start with the string "SFcn". For example, instead of naming your
     *      structure "ParseTree" you should name it "SFcnParseTree".
     *
     * Returns:
     *      1 -on success
     *      0 -on failure
     */
    bool WriteMatlabStructureToRTWFile(SimStruct      *S,
                                       const mxArray  *mlStruct,
                                       const char     *structName,
                                       char           *strBuffer,
                                       const int      strBufferLen)
    {
        int numStructs;
        int numFields;
        int i;
    
        if (mlStruct == NULL) {
            return(1);
        }
        numStructs = mxGetNumberOfElements(mlStruct);
        numFields  = mxGetNumberOfFields(mlStruct);
    
        /* make sure strBuffer is long enough for sprintf, be conservative */
        if ( strlen(structName)+5 >= ((size_t) strBufferLen) ) {
            return(0);
        }
    
        for (i=0; i<numStructs; i++) {
            int  j;
      
            (void) sprintf(strBuffer, "%s {",structName);
            if ( !ssWriteRTWStr(S, strBuffer) ) {
                return(0);
            }
      
            for (j=0; j<numFields; j++) {
                const char    *fieldName = mxGetFieldNameByNumber(mlStruct, j);
                const mxArray *field     = mxGetFieldByNumber(mlStruct, i, j);
                int           nRows;
                int           nCols;
                int           nElements;
                int           nDims;
	
                if (field == NULL) {
                    continue;
                }
	
                nRows      = mxGetM(field);
                nCols      = mxGetN(field);
                nElements  = mxGetNumberOfElements(field);
                nDims      = mxGetNumberOfDimensions(field);
	
                if ( mxIsStruct(field) ) {                       /* struct param */
                    if ( !WriteMatlabStructureToRTWFile(S,
                                                        field,
                                                        fieldName,
                                                        strBuffer,
                                                        strBufferLen) ) {
                        return(0);
                    }
	  
                } else if ( mxIsChar(field) ) {                  /* string param */
	  
                    /* can handle only "row" strings */
                    if ( nDims > 2 || nRows > 1 ) {
                        return(0);
                    }
	  
                    if ( mxGetString(field,strBuffer,strBufferLen) ) {
                        return(0);
                    }
	  
                    if ( !ssWriteRTWStrParam(S,fieldName,strBuffer) ) {
                        return(0);
                    }
	  
                } else if ( mxIsNumeric(field) ) {              /* numeric param */
	  
                    const void *rval   = mxGetData(field);
                    int        isCmplx = mxIsComplex(field);
                    DTypeId    dTypeId = ssGetDTypeIdFromMxArray(field);
                    int        dtInfo  = DTINFO(dTypeId, isCmplx);
                    const void *ival   = (isCmplx) ? mxGetImagData(field) : NULL;
	  
                    /* can handle only non-sparse, numeric vectors or matrices */
                    if (nDims > 2 || nRows*nCols != nElements || mxIsSparse(field)){
                        return(0);
                    }
	  
                    if (nRows == 1 || nCols == 1) {              /* vector param */
                        if ( !ssWriteRTWMxVectParam(S, fieldName, rval, ival,
                                                    dtInfo, nElements) ) {
                            return(0);
                        }
                    } else {                                     /* matrix param */
                        if ( !ssWriteRTWMx2dMatParam(S, fieldName, rval, ival,
                                                     dtInfo, nRows, nCols) ) {
                            return(0);
                        }
                    }
                } else {
                    return(0);
                }
            }
      
            if ( !ssWriteRTWStr(S, "}") ) {
                return(0);
            }
        }
        return(1);
    
    } /* end WriteMatlabStructureToRTWFile */
  


    int setSignalInfo(SimStruct *S,mxArray *sinfo, cs_signal_t **signals,int nosignals)
    {
        /*
          Returns 0 for success
          Return 1 for Failure
        */
        int i;
        uint16_t bit_len;
        uint16_t offset;
        cs_signal_byteorder_t bt;
        cs_signal_type_t signedness;
        double scale;
        double doffset;
        double minvalue;
        double maxvalue;
        double helpvalue;
    
    
        for(i = 0; i< nosignals; i++) {
            cs_signal_reset(signals[i]);
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"bit_start", &helpvalue);
            if((helpvalue < 0) || (helpvalue >63)) {
                ssSetErrorStatus(S, "Check Bit start value of signal");
                return 1;
            }
            offset = (uint16_t)helpvalue;
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"bit_len", &helpvalue);
            if((helpvalue < 1) || (helpvalue >63)) {
                ssSetErrorStatus(S, "Check Bit length value of signal");
                return 1;
            }
            bit_len = (uint16_t)helpvalue;
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"endianess", &helpvalue);
            if((helpvalue < 0) || (helpvalue >1)) {
                ssSetErrorStatus(S, "Check endianess of signal");
                return 1;
            }
            bt = (cs_signal_byteorder_t)helpvalue;
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"signedness", &helpvalue);
            if((helpvalue < 0) || (helpvalue >1)) {
                ssSetErrorStatus(S, "Check signedness of signal");
                return 1;
            }
            signedness = (cs_signal_type_t)helpvalue;
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"scale", &scale);
            if(scale == 0.0) {
                ssSetErrorStatus(S, "Check value scale of signal");
                return 1;
            }
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"offset", &doffset);
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"min", &minvalue);
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"max", &maxvalue);
            if(minvalue > maxvalue) {
                ssSetErrorStatus(S, "Check values min max of signal");
                return 1;
            }
      

	    if(cs_signal_set_signal_type(signals[i],signedness) ||
	       cs_signal_set_byteorder(signals[i],bt)||
	       cs_signal_set_bitoffset(signals[i],offset) ||
	       cs_signal_set_bitlength(signals[i],bit_len) ||
	       cs_signal_set_scale(signals[i],scale) ||
	       cs_signal_set_offset(signals[i],doffset) ||
	       cs_signal_set_minmax(signals[i],minvalue,maxvalue)){
                ssSetErrorStatus(S, "Failed to configure signal");
                return 1;
            }
        }
        return 0;
    }

    int checkSignalInfo(SimStruct *S,mxArray *sinfo)
    {
        /*
          Returns 0 for success
          Return 1 for Failure
        */
        int i;
        double helpvalue;
        double minvalue;
        double maxvalue;
        int nosignals = mxGetNumberOfElements(sinfo);
        
        
        for(i = 0; i< nosignals; i++) {
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"bit_start", &helpvalue);
            if((helpvalue < 0) || (helpvalue >63)) {
                ssSetErrorStatus(S, "Check Bit start value of signal");
                return 1;
            }
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"bit_len", &helpvalue);
            if((helpvalue < 1) || (helpvalue >63)) {
                ssSetErrorStatus(S, "Check Bit length value of signal");
                return 1;
            }
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"endianess", &helpvalue);
            if((helpvalue < 0) || (helpvalue >1)) {
                ssSetErrorStatus(S, "Check endianess of signal");
                return 1;
            }
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"signedness", &helpvalue);
            if((helpvalue < 0) || (helpvalue >1)) {
                ssSetErrorStatus(S, "Check signedness of signal");
                return 1;
            }
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"scale", &helpvalue);
            if(helpvalue == 0.0) {
                ssSetErrorStatus(S, "Check value scale of signal");
                return 1;
            }
      
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"offset", &helpvalue);
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"min", &minvalue);
            CHECK_STRUCT_VALUE_RVALUE( sinfo, i,"max", &maxvalue);
            if(minvalue > maxvalue) {
                ssSetErrorStatus(S, "Check values min max of signal");
                return 1;
            }
        }
        
        return 0;
    }
        
#ifdef __cplusplus
}
#endif

#endif
