/*  Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define S_FUNCTION_NAME  sfun_canrestbus_raw_read
#define S_FUNCTION_LEVEL 2

#include <inttypes.h>

#include "simstruc.h"

#include <restbus/cs_signal.h>
#include <restbus/cs_socketcan.h>

#include "sfun_canrestbus.h"


/*Defines*/

/*Parameter*/

#define PARAM_COUNT 6

#define PARAM_IFACE 0
#define PARAM_MINFO 1
#define PARAM_SINFO 2
#define PARAM_TIMEOUT 3
#define PARAM_DEBUG 4
#define PARAM_TSAMPLE 5

#define PARAM_ARG(PARAM)  (ssGetSFcnParam(S, PARAM))

/*Static outputs*/
#define OUT_TS 0
#define OUT_TIMEOUT 1
#define OUT_FAILURECOUNTER 2
#define OUT_FRAMECOUNTER 3
/*Number of static ports*/
#define OUT_STATIC 4


/*====================*
 * S-function methods 
 *====================*/

/* Function: mdlInitializeSizes =============================================
 * Abstract:
 *    The sizes information is used by Simulink to determine the S-function
 *    block's characteristics (number of inputs, outputs, states, etc.).
 */
static void mdlInitializeSizes(SimStruct *S)
{
   int_T i;
   double helpvalue;
   int nosignals;
   mxArray *sinfo = NULL;
   mxArray *minfo = NULL;


   ssSetNumSFcnParams(S, PARAM_COUNT);  /* Number of expected parameters */
   if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) {
     /* Return if number of expected != number of actual parameters */
     ssSetErrorStatus(S, "Parameter Failure");
     return;
   }
   
   for( i = 0; i < PARAM_COUNT; i++)
     ssSetSFcnParamTunable(S, i, SS_PRM_NOT_TUNABLE);
   
   
   /* Process input ports */
   if (!ssSetNumInputPorts(S, 0)) {
     ssSetErrorStatus(S, "Cannot set number of Input channels");
     return;
   }
   if(mxIsEmpty(PARAM_ARG(PARAM_IFACE))){
     ssSetErrorStatus(S, "No valid Interface spezified");
     return;
   }

   if(mxIsEmpty(PARAM_ARG(PARAM_TIMEOUT))){
     ssSetErrorStatus(S, "No valid timeout spezified");
     return;
   }
   
   if(mxIsEmpty(PARAM_ARG(PARAM_MINFO))){
     ssSetErrorStatus(S, "No valid Message structure defined");
     return;
   }
   
   minfo = (mxArray *)PARAM_ARG(PARAM_MINFO);
   if((minfo == NULL) || (mxGetNumberOfElements(minfo)!=1)){
     ssSetErrorStatus(S, "No valid Message structure defined or message ID  not found");
     return;
   }
   
   if(mxGetField(minfo, 0, "id")==NULL) {
       /*Skip if there is no valid structure*/
       nosignals = 0;
   } else {
     CHECK_STRUCT_VALUE( minfo, 0,"id", &helpvalue);
     if(helpvalue <= 0){
       ssSetErrorStatus(S, "Check message id");
       return;
     }
     CHECK_STRUCT_VALUE( minfo, 0,"dlc", &helpvalue);
     if((helpvalue <= 0) || (helpvalue > MAXDLC)) {
       ssSetErrorStatus(S, "Check message dlc");
       return;
     }
     
     sinfo = (mxArray *)PARAM_ARG(PARAM_SINFO);
     nosignals = mxGetNumberOfElements(sinfo);
   }
   

   /*ssPrintf("%s: Number of signals: %d\n",ssGetPath(S),nosignals);*/
   
   if (!ssSetNumOutputPorts(S, OUT_STATIC + nosignals)) {
     ssSetErrorStatus(S, "Cannot set number of Output channels");
     return;
   }

   /*Receive timestamp*/
   ssSetOutputPortWidth( S, OUT_TS, 1);
   ssSetOutputPortDataType( S, OUT_TS, SS_DOUBLE);

   /*Receive timeout flag*/
   ssSetOutputPortWidth( S, OUT_TIMEOUT, 1);
   ssSetOutputPortDataType( S, OUT_TIMEOUT, SS_BOOLEAN);

   /*Failure counter*/
   ssSetOutputPortWidth( S, OUT_FAILURECOUNTER, 1);
   ssSetOutputPortDataType( S, OUT_FAILURECOUNTER, SS_UINT32);

   /*Frame counter*/
   ssSetOutputPortWidth( S, OUT_FRAMECOUNTER, 1);
   ssSetOutputPortDataType( S, OUT_FRAMECOUNTER, SS_UINT32);
   
   
   for(i=0;i< nosignals;i++) {
     ssSetOutputPortWidth( S, OUT_STATIC + i, 1);
     ssSetOutputPortDataType( S, OUT_STATIC + i, SS_DOUBLE);
   }
   
   /*Transfer structure information*/
   if(checkSignalInfo(S,sinfo)){
     return;
   }

   

   ssSetNumContStates(S, 0);
   ssSetNumDiscStates(S, 0);
   ssSetNumSampleTimes(S, 1);
   ssSetNumRWork(S, 0);
   ssSetNumIWork(S, 0);
   ssSetNumPWork(S, 0);
   ssSetNumModes(S, 0);
   ssSetNumNonsampledZCs(S, 0);
   
   /* use generated code in Accelerator Mode */
   ssSetOptions(S, SS_OPTION_USE_TLC_WITH_ACCELERATOR
		| SS_OPTION_CALL_TERMINATE_ON_EXIT);
   
   /*ssSetOptions(S,
     SS_OPTION_WORKS_WITH_CODE_REUSE
     | SS_OPTION_RUNTIME_EXCEPTION_FREE_CODE
     | SS_OPTION_CALL_TERMINATE_ON_EXIT);
   */
}
  
/* Function: mdlInitializeSampleTimes =======================================
 * Abstract:
 *    This function is used to specify the sample time(s) for your
 *    S-function. You must register the same number of sample times as
 *    specified in ssSetNumSampleTimes.
 */
static void mdlInitializeSampleTimes(SimStruct *S)
{
  ssSetSampleTime(S, 0, mxGetScalar(ssGetSFcnParam(S, PARAM_TSAMPLE)));
  ssSetOffsetTime(S, 0, 0.0);
}

#define MDL_START                      /* Change to #undef to remove function */
#if defined(MDL_START)
/* Function: mdlStart ==========================================================
 * Abstract:
 *      Here we cache the state (true/false) of the XDATAEVENLYSPACED parameter.
 *      We do this primarily to illustrate how to "cache" parameter values (or
 *      information which is computed from parameter values) which do not change
 *      for the duration of the simulation (or in the generated code). In this
 *      case, rather than repeated calls to mxGetPr, we save the state once.
 *      This results in a slight increase in performance.
 */
static void mdlStart(SimStruct *S)
{
    int nosignals;
    double * oport;
    uint8_t * bport;
    uint32_t * iport;
    int i;
    
    nosignals = mxGetNumberOfElements((mxArray *)PARAM_ARG(PARAM_SINFO));
    
    if(nosignals <= 0){
        ssSetErrorStatus(S, "No signals defined");
        return;
    }
    
    /*Reset outputs*/
    
    /*Receice timestamp*/
    oport = (double *)ssGetOutputPortSignal(S,OUT_TS);
    oport[0] = 0.0;
    
    /*receive timeout flag*/
    bport = (uint8_t *)ssGetOutputPortSignal(S,OUT_TIMEOUT);
    bport[0] = 0; /*No timeout*/
    
    /*Failure counter*/
    iport = (uint32_t *)ssGetOutputPortSignal(S,OUT_FAILURECOUNTER);
    iport[0] = 0;
    
    /*Frame counter*/
    iport = (uint32_t *)ssGetOutputPortSignal(S,OUT_FRAMECOUNTER);
    iport[0] = 0;
    
    for(i = 0; i < nosignals; i++) {
        oport = (double *)ssGetOutputPortSignal(S,OUT_STATIC + i);
        oport[0] = 0.0;
    }
    
}
#endif /*  MDL_START */



#define MDL_UPDATE
static void mdlUpdate(SimStruct *S, int_T tid)
{
  /** Needed for Matlab versions from 2007b, otherwise no update code is
   * generated.
   */
}

/* Function: mdlOutputs =====================================================
 * Abstract:
 *    In this function, you compute the outputs of your S-function
 *    block. Generally outputs are placed in the output vector, ssGetY(S).
 */
static void mdlOutputs(SimStruct *S, int_T tid)
{

  
}

/* Function: mdlTerminate ===================================================
 * Abstract:
 *    In this function, you should perform any actions that are necessary
 *    at the termination of a simulation.  For example, if memory was
 *    allocated in mdlStart, this is the place to free it.
 */
static void mdlTerminate(SimStruct *S)
{

}


#define MDL_RTW  /* Change to #undef to remove function */
#if defined(MDL_RTW) && (defined(MATLAB_MEX_FILE) || defined(NRT))
static void mdlRTW(SimStruct *S){
  double timeout_s = mxGetScalar(PARAM_ARG(PARAM_TIMEOUT));
  char *iface = mxArrayToString(PARAM_ARG(PARAM_IFACE));
  int buflen = 1500;
  char buffer[buflen];
  mxArray *minfo = (mxArray *)PARAM_ARG(PARAM_MINFO);
  mxArray *sinfo = (mxArray *)PARAM_ARG(PARAM_SINFO);
  
  if((minfo == NULL) ||
     (sinfo == NULL) ||
     (mxGetField(minfo, 0, "id")==NULL) ||
     (mxGetNumberOfElements(sinfo) <=0)
     ) {
    ssSetErrorStatus(S,"Check inital data structures.\n");
    return;
  }


  /* -- Write Invariant Parameter Settings -- */
   ssWriteRTWParamSettings(S,2,
			   SSWRITE_VALUE_DTYPE_NUM, "TIMEOUT", &timeout_s, DTINFO(SS_DOUBLE,0),
			   SSWRITE_VALUE_QSTR,"IFACE",iface
			   );

   if(!WriteMatlabStructureToRTWFile(S,
				     ssGetSFcnParam(S,PARAM_SINFO),
				     "SINFO",
				     buffer, buflen)){
     ssSetErrorStatus(S,"Writing MATLAB structure to Real-Time Workshop file failed.\n");
     return;
   };

   if(!WriteMatlabStructureToRTWFile(S,
				     ssGetSFcnParam(S,PARAM_MINFO),
				     "MINFO",
				     buffer, buflen)){
     ssSetErrorStatus(S,"Writing MATLAB structure to Real-Time Workshop file failed.\n");
      return;
   };
}
#endif

/*======================================================*
 * See sfuntmpl_doc.c for the optional S-function methods *
 *======================================================*/

/*=============================*
 * Required S-function trailer *
 *=============================*/

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
