/*  Copyright (C) 2018 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#define S_FUNCTION_NAME  sfun_canrestbus_sleep_rt
#define S_FUNCTION_LEVEL 2


#include "simstruc.h"

#include <time.h>

/*Parameter*/

#define PARAM_COUNT 1

#define TSAMPLE 0
#define PARAM_ARG(PARAM)  (ssGetSFcnParam(slave->S, PARAM))


#define CLOCK CLOCK_MONOTONIC
#define TIMER TIMER_ABSTIME
#define NSEC_PER_SEC (1000000000L)


struct cs_canrestbus_realtime {
  uint_T overruns;
  struct timespec lastcall;
  struct timespec wakeup;
  struct timespec difftime;
  struct timespec lastwakeup;
};


struct timespec ts_addtimespec(struct timespec time1, struct timespec time2)
{
    struct timespec result;
    if((time1.tv_nsec + time2.tv_nsec) >= 1e9) {
      result.tv_sec = time1.tv_sec + time2.tv_sec + 1;
      result.tv_nsec = time1.tv_nsec +time2.tv_nsec -1e9;
    } else {
      result.tv_sec = time1.tv_sec + time2.tv_sec ;
      result.tv_nsec = time1.tv_nsec +time2.tv_nsec;
    }
    return result;
}

struct timespec ts_difftimespec(struct timespec start, struct timespec end)
{
  struct timespec temp;
  if ((end.tv_nsec-start.tv_nsec)<0) {
    temp.tv_sec = end.tv_sec-start.tv_sec-1;
    temp.tv_nsec = 1000000000+end.tv_nsec-start.tv_nsec;
  } else {
    temp.tv_sec = end.tv_sec-start.tv_sec;
    temp.tv_nsec = end.tv_nsec-start.tv_nsec;
  }
  return temp;
}

/*====================*
 * S-function methods *
 *====================*/

/* Function: mdlInitializeSizes =============================================
 * Abstract:
 *    The sizes information is used by Simulink to determine the S-function
 *    block's characteristics (number of inputs, outputs, states, etc.).
 */
static void mdlInitializeSizes(SimStruct *S)
{
    uint_T i;
    
    ssSetNumSFcnParams(S, PARAM_COUNT);  /* Number of expected parameters */
    if (ssGetNumSFcnParams(S) != ssGetSFcnParamsCount(S)) {
        /* Return if number of expected != number of actual parameters */
        ssSetErrorStatus(S, "Parameter Failure");
        return;
    }
    
    for( i = 0; i < PARAM_COUNT; i++)
        ssSetSFcnParamTunable(S, i, SS_PRM_NOT_TUNABLE); 

    /* Process input ports */
    if (!ssSetNumInputPorts(S, 0))
        {
            ssSetErrorStatus(S, "Cannot set number of Input channels");
            return;
        }
    if (!ssSetNumOutputPorts(S, 3))
        {
            ssSetErrorStatus(S, "Cannot set number of Output channels");
            return;
        }
    /*Overruns*/
    ssSetOutputPortWidth   (S, 0, 1);
    ssSetOutputPortDataType(S, 0, SS_UINT32);
    /*Jitter*/
    ssSetOutputPortWidth   (S, 1, 1);
    ssSetOutputPortDataType(S, 1, SS_UINT32);
    /*Period*/
    ssSetOutputPortWidth   (S, 2, 1);
    ssSetOutputPortDataType(S, 2, SS_UINT32);

    ssSetNumContStates(S, 0);
    ssSetNumDiscStates(S, 0);
    ssSetNumSampleTimes(S, 1);
    ssSetNumRWork(S, 0);
    ssSetNumIWork(S, 0);
    ssSetNumPWork(S, 0);
    ssSetNumModes(S, 0);
    ssSetNumNonsampledZCs(S, 0);

    /* Set the options. Note that SS_OPTION_EXCEPTION_FREE_CODE cannot
     * be set because this SFunction uses mxCalloc, which can cause
     * an exception */
    ssSetOptions(S,
                 SS_OPTION_WORKS_WITH_CODE_REUSE
                 | SS_OPTION_RUNTIME_EXCEPTION_FREE_CODE
                 | SS_OPTION_CALL_TERMINATE_ON_EXIT);
}

/* Function: mdlInitializeSampleTimes =======================================
 * Abstract:
 *    This function is used to specify the sample time(s) for your
 *    S-function. You must register the same number of sample times as
 *    specified in ssSetNumSampleTimes.
 */
static void mdlInitializeSampleTimes(SimStruct *S)
{
    ssSetSampleTime(S, 0, mxGetScalar(ssGetSFcnParam(S, TSAMPLE)));
    ssSetOffsetTime(S, 0, 0.0);
}


#define MDL_START                      /* Change to #undef to remove function */
#if defined(MDL_START)
/* Function: mdlStart ==========================================================
 */
static void mdlStart(SimStruct *S)
{
  struct cs_canrestbus_realtime *slave_realtime;
  /* allocate memory for slave structure */
  if (!(slave_realtime = malloc(sizeof(struct cs_canrestbus_realtime)))) {
    ssSetErrorStatus(S, "Malloc Failure");
    return;
  }
  slave_realtime->overruns = 0;
  slave_realtime->difftime.tv_sec = 0;
  slave_realtime->difftime.tv_nsec = ((long)(mxGetScalar(ssGetSFcnParam(S, TSAMPLE))*NSEC_PER_SEC));
  clock_gettime(CLOCK, &slave_realtime->lastcall); 
  slave_realtime->lastwakeup = slave_realtime->lastcall;
  slave_realtime->wakeup = ts_addtimespec(slave_realtime->lastcall,slave_realtime->difftime);
  
  ssSetUserData(S,slave_realtime);
 }
#endif /*  MDL_START */
 
#define MDL_UPDATE
 static void mdlUpdate(SimStruct *S, int_T tid)
 {
  /** Needed for Matlab versions from 2007b, otherwise no update code is
   * generated.
   */
 }
 
 /* Function: mdlOutputs =====================================================
  * Abstract:
  *    In this function, you compute the outputs of your S-function
  *    block. Generally outputs are placed in the output vector, ssGetY(S).
  */
 static void mdlOutputs(SimStruct *S, int_T tid)
 {
   uint32_T *datap_overruns = NULL;
   uint32_T *datap_jitter = NULL;
   uint32_T *datap_period = NULL;
   struct cs_canrestbus_realtime *slave_realtime = ssGetUserData(S);
   struct timespec acttime;
   struct timespec jittertime;
   struct timespec periodtime;
   
   datap_overruns = (uint32_T *)ssGetOutputPortSignal(S,0);
   datap_jitter = (uint32_T *)ssGetOutputPortSignal(S,1);
   datap_period = (uint32_T *)ssGetOutputPortSignal(S,2);
   
   if(clock_nanosleep(CLOCK,TIMER,&slave_realtime->wakeup,NULL)!=0) {
     slave_realtime->overruns++;
   }
   clock_gettime(CLOCK,&acttime);
   jittertime = ts_difftimespec(slave_realtime->wakeup, acttime);
   if(jittertime.tv_nsec > slave_realtime->difftime.tv_nsec) {
     slave_realtime->overruns++;
   }
   periodtime = ts_difftimespec(slave_realtime->lastwakeup, acttime);
   slave_realtime->lastwakeup = acttime;
   datap_overruns[0] = slave_realtime->overruns;
   datap_jitter[0] = (uint32_T)jittertime.tv_nsec;
   datap_period[0] = (uint32_T)periodtime.tv_nsec;
   slave_realtime->lastcall = slave_realtime->wakeup;
   slave_realtime->wakeup = ts_addtimespec(slave_realtime->lastcall,slave_realtime->difftime);
}

/* Function: mdlTerminate ===================================================
 * Abstract:
 *    In this function, you should perform any actions that are necessary
 *    at the termination of a simulation.  For example, if memory was
 *    allocated in mdlStart, this is the place to free it.
 */
static void mdlTerminate(SimStruct *S)
{
  struct cs_canrestbus_realtime *slave_realtime = ssGetUserData(S);

  if(slave_realtime) {
    free(slave_realtime);
  }
}

/*======================================================*
 * See sfuntmpl_doc.c for the optional S-function methods *
 *======================================================*/

/*=============================*
 * Required S-function trailer *
 *=============================*/

#ifdef  MATLAB_MEX_FILE    /* Is this file being compiled as a MEX-file? */
#include "simulink.c"      /* MEX-file interface mechanism */
#else
#include "cg_sfun.h"       /* Code generation registration function */
#endif
